import ch.epfl.lara.rmf._

/**
 * An object of this class represent a TPTP Problem
 * Translated to kodkod
 */
class KodKodProblem {
  
  /**
   * The Bounds of this instance (the domain we're looking for a model in)
   **/
  var bounds: kodkod.instance.Bounds = null
    
  /**
   * Formulas that have to be satisfied
   */
  var declarations: List[kodkod.ast.Formula] = Nil
  
  /**
   * Conterexamples have to be found
   */
  var conjectures: List[kodkod.ast.Formula] = Nil
  

  /**
   * Context is a list of leaf expressions: relations, predicates, constants, ...
   */
  type Context = List[kodkod.ast.LeafExpression]
                      
  /**
   * The context of our problem: relations, predicates, constants.
   */
  var globalExpressions: Context = Nil
  
  /**
   * Formulas that will be solved by kodkod
   * Conjunction of declarations and negation of conjecture
   */
  var formulas: List[kodkod.ast.Formula] = null
  
  /**
   * Construct this problem from a list of annoted tptp statements
   */
  def this(dSize:Int, afs:List[TPTPExpression]) =
  {
    this()
    // Create domain
    val jc = new java.util.LinkedList()
    (1 until dSize+1) foreach (i => jc.add("E"+i))
    bounds = new kodkod.instance.Bounds(new kodkod.instance.Universe(jc))
    // Gather all relations
    buildInitialContext(afs)
    //Translate to KodKod
    translateToKodKod(afs)
    //Assemble axioms and conjectures
    buildCompleteFormula
  }
  
  /**
   * Construct a problem directly from file
   */
  def this(dSize:Int,pbDir:String, incDir:String, pbFile:String) =
  {
    this(dSize,TPTPParser.parse(pbDir,incDir,pbFile))
  }
  
  /**
   * Solve this problem with the given SAT Solver
   */
  def solve(solver:kodkod.engine.Solver):Unit=
  {    
    formulas foreach (f =>
            {
              //Console.println(f)
              val sol = solver.solve(f, bounds)
              Console.println(sol)
            })
 }
  
  /**
   * Build the conjunctions of declarations and the negation of conjecture
   */
  def buildCompleteFormula : Unit = 
  {
    val decls = declarations reduceLeft ( (l:kodkod.ast.Formula,r:kodkod.ast.Formula) => l.and(r))
    formulas = conjectures map (c => decls.and(c.not()))
  }
  
  
  /**
   * Methods for translating TPTP Element to KodKod
   */
  
  /**
   * Translate a list of annoted formulas to KodKod representation (stored in the adequate place)
   */
  def translateToKodKod(as: List[TPTPExpression]): Unit = as foreach translateToKodKod
 
  /**
   * Translate an annoted formulas to KodKod representation (stored in the adequate place)
   */
  def translateToKodKod(a: TPTPExpression): Unit = a match
  {
    case AnnotedFormula(name,Axiom,f) => declarations ::= translateToKodKod(f)
    case AnnotedFormula(name,Theorem,f) => declarations ::= translateToKodKod(f)
    case AnnotedFormula(name,Lemma,f) => declarations ::= translateToKodKod(f)
    case AnnotedFormula(name,Hypothesis,f) => declarations ::= translateToKodKod(f)
    
    case AnnotedFormula(name,Conjecture,f) => conjectures ::= translateToKodKod(f)
    
    case AnnotedFormula(name,role,f) => throw new IllegalArgumentException(role + " is not supported.")
    
    case AnnotedClause(name,role,f) => throw new UnsupportedOperationException()
    
    case _ => throw new IllegalArgumentException("AnnotedElement required")
  }
  
  /**
   * Translate a TPTP Formula to a KodKod Formula
   */
  def translateToKodKod(f: Formula):kodkod.ast.Formula = f match
      {
      case Negation(f) => translateToKodKod(f).not()
      case And(l, r) => translateToKodKod(l).and(translateToKodKod(r))
      case Or(l, r) => translateToKodKod(l).or(translateToKodKod(r))
      case Atomic(predName, args) if(predName == "=" && args.size == 2) =>{
        val as:List[kodkod.ast.Expression] = args map (a => translateToKodKod(a))
        as(0).eq(as(1))
      }
      case Atomic(predName, args) if(predName == "=" && args.size != 2) => throw new RuntimeException("Predicate '=' called with more than two arguments")
      case Atomic(predName, args) if(predName != "=") => (args map (a => translateToKodKod(a)) reduceLeft ((l:kodkod.ast.Expression,r:kodkod.ast.Expression) => l.product(r))).in(findRelationInContext(predName))
      
      case Implication(l, r) => translateToKodKod(l).implies(translateToKodKod(r))
      case Equivalence(l, r) => translateToKodKod(l).iff(translateToKodKod(r))
      case ReverseImplication(l, r) => translateToKodKod(Implication(r,l))
      case Disequivalence(l, r) => translateToKodKod(Equivalence(l,r)).not()
      case NotOr(l, r) => translateToKodKod(Or(l,r)).not()
      case NotAnd(l, r) => translateToKodKod(And(l,r)).not()
      case Exists(varName, f) => 
      {
        val v = addVariable(varName)
        val qf = translateToKodKod(f).forSome(v.oneOf(kodkod.ast.Expression.UNIV))
        popVariable()
        qf
      }
      case ForAll(varName, f) => 
      {
        val v = addVariable(varName)
        val qf = translateToKodKod(f).forAll(v.oneOf(kodkod.ast.Expression.UNIV))
        popVariable()
        qf
      }
      }

  /**
   * Translate a TPTP Term to KodKod Expression
   */
  def translateToKodKod(t: Term):kodkod.ast.Expression = t match
  {
  case Term(Symbol(name,isVar),args) if(isVar) => findVariableInContext(name)
  case Term(Symbol(name,isVar),args) if(!isVar && args.size == 0) => findRelationInContext(name)
  case Term(Symbol(name,isVar),args) if(!isVar && args.size != 0) =>
          {
            val kodkodArgs =(args map (a => translateToKodKod(a))) reverse
            val rel = findRelationInContext(name)
            
            def application(args:List[kodkod.ast.Expression]):kodkod.ast.Expression = args match {
              case Nil=> rel;
              case x::xs => x.join(application(xs))
            }
            
            application(kodkodArgs)
          }
  }
  
  /**
   * Methods for modifiying the symbols list
   * Symbols list keeps track of variables, relations, constants...
   */
     
  /**
   * Check whether a given name is in the symbols list
   */
  def isInContext(name:String):Boolean = globalExpressions.find(l => l.name() == name) match
  {
  case Some(v) => true
  case None => false
  }
  
  /**
   * Find a given name is in the symbols list
   * @throw IllegalStateExcpetion if the requested name is not in the symbole list
   */
  def findInContext(name:String):kodkod.ast.LeafExpression = globalExpressions.find(l => l.name() == name) match
  {
  case Some(v) => v
  case None => throw new IllegalStateException("Element " + name + " not found in context.")
  }
  
  /**
   * Convenient method to find relations in symbols list
   */
  def findRelationInContext(name:String):kodkod.ast.Relation = 
  {
    val el = findInContext(name)
    if(el.isInstanceOf[kodkod.ast.Relation])
      el.asInstanceOf[kodkod.ast.Relation]
    else
      throw new IllegalStateException("Element " + name + " is not a relation.")
  }
  
  /**
   * Convenient method to find Variable in symbols list
   */
  def findVariableInContext(name:String):kodkod.ast.Variable = 
  {
    val el = findInContext(name)
    if(el.isInstanceOf[kodkod.ast.Variable])
      el.asInstanceOf[kodkod.ast.Variable]
    else
      throw new IllegalStateException("Element " + name + " is not a variable.")
  }
  
  /**
   * Convenient method to declare a new constant in symbols list
   */
  def addConstant(name:String):Unit = 
    if (!isInContext(name))
      {
        val rel = kodkod.ast.Relation.unary(name)
        val ts = bounds.universe().factory().allOf(1)
        bounds.bound(rel,ts)
        globalExpressions ::= rel
        declarations ::= rel.one()
      }
  
  /**
   * Convenient method to add a new relation in symbols list
   */
  def addRelation(name:String, n:Int):Unit =
    if (!isInContext(name)) 
    {
      val rel = kodkod.ast.Relation.nary(name,n)
      val ts = bounds.universe().factory().allOf(n)
      bounds.bound(rel,ts)
      globalExpressions ::= rel
    }
  
  /**
   * Convenient method to push a variable in symbols list
   */
  def addVariable(name:String):kodkod.ast.Variable =
  {
    val v = kodkod.ast.Variable.unary(name)
    globalExpressions ::= v
    v
  }
  
  /**
   * Remove the latest insterted variable from the symbols list
   * Used when a quantified formula is left
   */
  def popVariable():Unit = globalExpressions match
  {
    case x::xs if (x.isInstanceOf[kodkod.ast.Variable]) => globalExpressions = xs
    case x::xs if (!x.isInstanceOf[kodkod.ast.Variable]) => throw new IllegalStateException("Top of Symbol Table is not a variable.")
    case Nil => throw new IllegalStateException("Symbol Table is empty.")
  }
  
  /**
   * Looks in formulas for predicates, functions and constants
   */
    def buildInitialContext(as:List[TPTPExpression]):Unit = as match
    {
    case expr::rest => 
    {
      buildInitialContext(expr)
      buildInitialContext(rest)
    }
    case Nil => 
    }
    
    /**
     * Looks in formula for predicates, functions and constants
     */
    def buildInitialContext(expr: TPTPExpression):Unit = expr match
    {
          case AnnotedClause(name, role, clause) => buildInitialContext(clause)
          case AnnotedFormula(name, role, formula) => buildInitialContext(formula)
          
          case Atomic(predName, args) if (predName == "=") => buildInitialContext(args)
          case Atomic(predName, args) if (predName != "=") => 
          {
            addRelation(predName,args.size)
            buildInitialContext(args)
          }
          case Negation(f) => buildInitialContext(f)
          case And(l, r) => buildInitialContext(l::r::Nil)
          case Or(l, r) => buildInitialContext(l::r::Nil)
          case Equivalence(l, r) => buildInitialContext(l::r::Nil)
          case Implication(l, r) => buildInitialContext(l::r::Nil)
          case ReverseImplication(l, r) => buildInitialContext(l::r::Nil)
          case Disequivalence(l, r) => buildInitialContext(l::r::Nil)
          case NotOr(l, r) => buildInitialContext(l::r::Nil)
          case NotAnd(l, r) => buildInitialContext(l::r::Nil)
          case Exists(varName, f) => buildInitialContext(f)
          case ForAll(varName, f) => buildInitialContext(f)
          case Term(Symbol(name, isVar), args) if(isVar) =>  buildInitialContext(args)
          case Term(Symbol(name, isVar), args) if(!isVar && args.size == 0) =>
          {
            addConstant(name)
            buildInitialContext(args)
          }
          case Term(Symbol(name, isVar), args) if(!isVar && args.size != 0) =>
          {
            addRelation(name,args.size+1)
            buildInitialContext(args)
          }
          case Clause(lits) => buildInitialContext(lits)
          case Literal(isPos, atome) => buildInitialContext(atome)
    }
}
