package ch.epfl.lara.rmf;
import java.io.PrintWriter

  /**
   * Pretty printer using scala.text formatting library. It works
   * by first building a document abstracting over nesting, spacing and
   * new lines, and afterwards rendering the text given a Writer and a
   * desired width.
   */
    object RMFPrettyPrinter extends ((TPTPExpression,java.io.PrintStream, Int)=>Unit) {
      def apply(t: TPTPExpression,out: java.io.PrintStream, width:Int) = {
        val writer = new PrintWriter(out)
        toDocument(t).format(width, writer)
        writer.println
        writer.flush()
      }
      
      import scala.text._
      import scala.text.Document._
      
      def toDocument(ts: List[TPTPExpression], sep: String, suffix: String): Document = ts match {
        case one :: two :: rest =>
          toDocument(one) :: sep :/: rest.foldLeft(toDocument(two))  { (d, e) => 
            if (sep != "") d :: sep :/: toDocument(e) 
            else d :/: toDocument(e) } :: text(suffix)
        case one :: Nil =>
          toDocument(one) :: text(suffix)
        case Nil =>
          empty
      }
          
      def toDocument(t: TPTPExpression): Document = t match {
        case AnnotedClause(name, role, clause) => group(text(name) :/: "(" :: role.toString :: text("):")  :/: nest(2, group(toDocument(clause)))) :/: empty
        case AnnotedFormula(name, role, formula) =>  group( text(name) :/: "(" :: role.toString :: text("):")  :/: nest(2, group(toDocument(formula)))) :/: empty
        case Clause(lits) => toDocument(lits,"|","")
        case Literal(positive,atom) => if(positive) toDocument(atom) else group("~" :: toDocument(atom))
        case Atomic(pred,args) => pred :: "(" :: group(toDocument(args, ",", "")) :: text(")") 
        case Negation(f) => "~"::toDocument(f)
        case Term(sym,args) => args match {
          case Nil => toDocument(sym)
          case _ => group(toDocument(sym)::group("("::toDocument(args,",",")")))
        }
        case Symbol(name,isVar) => if(isVar)text(name) else text("-"+name+"-")
        case And(l,r) => toDocument(l)::"&"::toDocument(r)
        case Or(l,r) => toDocument(l)::"|"::toDocument(r)
        case Equivalence(l,r) => toDocument(l)::"<=>"::toDocument(r)
        case Implication(l,r) => toDocument(l)::"=>"::toDocument(r)
        case ReverseImplication(l,r) => toDocument(l)::"<="::toDocument(r)
        case Disequivalence(l,r) => toDocument(l)::"<~>"::toDocument(r)
        case NotOr(l,r) => toDocument(l)::"~|"::toDocument(r) 
        case NotAnd(l,r) => toDocument(l)::"~&"::toDocument(r)
        case ForAll(name,f) => text("ForAll"):/:text(name):/:text("(")::toDocument(f)::text(")")
        case Exists(name,f) => text("Exists"):/:text(name):/:text("(")::toDocument(f)::text(")")
        case _ => text(t.toString)
      }
  }