package ch.epfl.lara.rmf.interpretation;

import scala.collection.mutable.HashMap;

/**
 * Environment of our interpretations ("Interpretation" was already taken)
 */    
class Environment extends Iterable {
 
  /** Domain we are working in */
  var domain:Domain = null;
  
  /** Iterator Stack to enumerate all possible interpretations */
  var symbolStack:IteratorStack = null;
  
  /** List of Function Symbols in our Formula */        
  var functions:List[FunctionRepr] = Nil;
  /** List of Predicate Symbols in our Formula */
  var predicates:List[PredicateRepr] = Nil;
  
  /** Hashmaps for quick research of symbols */
  var functionsHash:HashMap[String, FunctionRepr] = new HashMap();  
  var predicatesHash:HashMap[String, PredicateRepr] = new HashMap();   
  
  /** Constructor */
  def this(d:Domain, symbols:List[InterpretationElementRepr]) = { 
    this();
    this.domain = d;
    
    //build the symbol iterator stack
    for (val symbol <- symbols) {
      symbol.initInterpretationStack;
      symbolStack = new IteratorStack(symbol, symbolStack);
    }

    //split the functions and predicates
    for (val symbol <- symbols) {
      if (symbol.isInstanceOf[FunctionRepr]) {
        functions ::= symbol.asInstanceOf[FunctionRepr];
        functionsHash.update(symbol.name, symbol.asInstanceOf[FunctionRepr]);
      } else {
        predicates ::= symbol.asInstanceOf[PredicateRepr];
        predicatesHash.update(symbol.name, symbol.asInstanceOf[PredicateRepr]);
      }
    }
  }
  
  /** Lookup a function */
  def findFunction(name:String):FunctionRepr = functionsHash.get(name) match {
    case Some(f) => f;
    case None => null;
  }
  
  /** Lookup a predicate */
  def findPredicate(name:String):InterpretationElementRepr = predicatesHash.get(name) match {
    case _ if name == "=" => EqPredicate
    case Some(p) => p;
    case None => null;
  }   

  // ------------------------------------------------------------------------------------
  /** Iterable */
  def hasNext():Boolean = {
    return symbolStack.hasNext();
  }
  def next() = {
    symbolStack.next(); 
  }
  def reset() = {
    symbolStack.reset();
  }
  
  /** Randomize the configuration */
  def rand(rand:java.util.Random) = {
    symbolStack.rand(rand);
  }
  
  // ------------------------------------------------------------------------------------
  /*
   * Export this interpretation as a set of TPTP AnnotedFormula
   */
  def export:List[AnnotedFormula] = 
    domain.export :: (
      functions map (f => f.export)).toList ::: (predicates  map (p => p.export)).toList
  
}