  package ch.epfl.lara.rmf.interpretation;

/**
 * This class will perform the evaluation of a formula for a given interpretation
 */
object Evaluator {

  /** Context: Assignments of our variables to domain elements */
  type Context = List[Pair[String, Element]]

   def evaluate(as: List[TPTPExpression], env:Environment):Int = (as filter (a => !evaluate(a, env, 0.0))).size
                       
   def evaluate(a: TPTPExpression, env: Environment, speed:Double): Boolean = a match {
     case AnnotedFormula(name, role, f) => evaluate(f, env, Nil, speed)
     case AnnotedClause(name, role, f) => throw new UnsupportedOperationException()
     case _ => throw new IllegalArgumentException("AnnotedElement required")
   }
  
   def evaluate(f: Formula, env: Environment, ctx: Context, speed:Double):Boolean = f match {
     case Negation(f) => !(evaluate(f, env, ctx, speed ))
     case And(l, r) => evaluate(l, env, ctx, speed) && evaluate(r, env, ctx, speed)
     case Or(l, r) =>evaluate(l, env, ctx, speed) || evaluate(r, env, ctx, speed)

     /** Evaluation of Predicates */
     case Atomic(predName, args) => {
       var argValues = args map (a => evaluate(a, env, ctx, speed).id);
       var pred = env.findPredicate(predName);
       
       //propagate speeds
       //propagate speed changes if required
       if (speed > 0.1) { 
         pred.increaseSpeed(argValues);           
       } else if (speed < 0.1) {
         pred.decreaseSpeed(argValues);
       }
       
       //evaluate
       pred.evaluate(argValues) match {
         case 1 => true  
         case 2 => false
       }
     }
   
     case Implication(l, r) => evaluate(Or(Negation(l),r), env, ctx, speed)
     case Equivalence(l, r) => evaluate(And(Implication(l,r),Implication(r,l)), env, ctx, speed)
     case ReverseImplication(l, r) => evaluate(Implication(r,l), env, ctx, speed)
     case Disequivalence(l, r) => !evaluate(Equivalence(l,r), env, ctx, speed)
     case NotOr(l, r) => !(evaluate(Or(l,r), env, ctx, speed))
     case NotAnd(l, r) => !(evaluate(And(l,r), env, ctx, speed))
     case Exists(varName, f) => env.domain.elements exists (e => evaluate(f, env, Pair(varName,e)::ctx, speed))
     case ForAll(varName, f) => {
       env.domain.elements forall (e => {
         //evaluate the forall
         var p = evaluate(f, env, Pair(varName,e)::ctx, speed);
         
         //propagate speed changes
         if (!p) {
           evaluate(f, env, Pair(varName,e)::ctx, 1.0);
         } else {
           evaluate(f, env, Pair(varName,e)::ctx, -1.0);
         }
         
         //ret the predicate
         p;
       })
     }
   }  
   
  /** Evaluate a term */
  def evaluate(t: Term, env: Environment, ctx: Context, speed:Double):Element = t match {
     case Term(Symbol(name, isVar), args) => 
     if (isVar) {
       // looking in ctx
       ctx find ( c => c._1 == name) match {
         case Some(Pair(varName, element)) => element
         case None => throw new RuntimeException("Cannot find var " + name + " in context.")
       }
     } else {
       //evaluate arguments and lookup function
       var argValues = args map (a => evaluate(a, env, ctx, speed).id);
       var function = env.findFunction(name);

       //propagate speed changes if required
       if (speed > 0.1) { 
	 function.increaseSpeed(argValues);           
       } else if (speed < 0.1) {
         function.decreaseSpeed(argValues);
       }
       
       // evaluate the function
       return new Element(function.evaluate(argValues));
     }
   }       
  
}
