package ch.epfl.lara.rmf.interpretation;

/***
 * These classes are descriptors for objects whithin an interpretation
 */

/**
 * This class describes the interpretation space for a formula (domain and interpretable functions
 * and predicates)
 */   
case class InterpretationRepr(dSize:Int, fcts: List[FunctionRepr], preds: List[PredicateRepr]) {}

/**
 * Superclass of both function and predicate representations 
 */
trait InterpretationElementRepr extends Iterable {
   
   //forward declarations
   def domain:Domain;
   def name:String;
   def num:Int;
   def imageSize:Int;
	
   /** This stack will enumerate us all possible variables */  
   var varList:List[IterableVariable] = Nil;
   var interpretationStack:Iterable = null;
   
   /** Speed associated with this element [] */
   var speedList:List[Speed] = Nil;
   
   /** Initializes the interpretation stack */
   def initInterpretationStack:Unit = {
     //create the variables of the interpretation and connect them
     var size:Int = java.lang.Math.pow(domain.size, num).asInstanceOf[Int];  
     while (size > 0) {
       size -= 1;
       var temp:IterableVariable = new IterableVariable(imageSize);
       varList = temp :: varList;
       interpretationStack = new IteratorStack(temp, interpretationStack);
       
       //initialize speed
       speedList = new Speed() :: speedList;
     }   
   };
   
   /** Increase the speed of modifications for this entry in the interpretation */
   def increaseSpeed(pars:List[Int]):Unit = {
     var index = toIndex(pars);
     speedList(index).increase();
   }
   /** Lower the speed of modifications for this entry in the interpretation */
   def decreaseSpeed(pars:List[Int]):Unit = {
     var index = toIndex(pars);
     //Console.println("Decreasing Speed");
     //for (i <- pars) Console.print("  " + i);
     //Console.println(" -> " + index);
     speedList(index).decrease();
   }

   /** Convert a set of parameter values to the index of the interpretation */
   def toIndex(pars:List[Int]):Int = {
     //each combination of input values is represented by an iterator
     //so now we have to look it up
     var offset:Int = 0;
     var index:Int = 1;
     while (index <= num) {
       //read the selector value
       var value:Int = pars(index - 1) - 1;

       //adjust offset
       offset += (value * java.lang.Math.pow(domain.size, num - index).asInstanceOf[Int]);
       index += 1;
     }
     
     //for (p <- pars) Console.print(p + " ");
     //Console.println(" -> " + offset);
     return offset;
   }
     
   /** Reset this symbol */
   def reset():Unit = {
     //delegate the reset to our interpretation stack
     interpretationStack.reset();
   }

   /** Check if more interpretations are avaiable */
   def hasNext():Boolean =  {
      return interpretationStack.hasNext();
   }
       
   /** Set the internal state to the next interpretation */
   def next():Unit = {
     interpretationStack.next();
   }
       
   /** Randomize the configuration */
   def rand(rand:java.util.Random) = {
     //forward rand request to selected elements
     (speedList zip varList) foreach (pair => if (pair._1.shouldChange(rand)) {
       pair._2.rand(rand);
     })   
     //	interpretationStack.rand(rand);
   }
   
   /**
    * Evaluate the current interpretation with a set of real argument values
    */
   def evaluate(parameter:List[Int]):Int;   
}

/**
 * Interpretable functions
 *
 * This describes a function of arity num that evaluates within our domain
 */
case class FunctionRepr(domain:Domain, name:String, num:Int) extends InterpretationElementRepr() {
  /** Functions can evaluate to all domain elements */	
  def imageSize = domain.size;
  
  override def toString():String = "Function " + name + " (arity: " + num + ")";
  
  /**
   * Evaluate the current interpretation with a set of real argument values
   */
  def evaluate(parameter:List[Int]):Int = {
    return varList(toIndex(parameter)).getValue();
  }
   
  /** Enumerate the interpretation */
  def getDefinedEntries():List[Pair[List[Element], Element]] = {
    var buffer:List[Pair[List[Element],Element]] = Nil; 
    var index:Int = 0;
    while (index < varList.size) {
      //restore arguments        
      var pars:List[Element] = Nil;
        
      //now we have the pleasure of converting the index into the parameters
      var value = index; //base value
      var varIdx = 1;
      while (varIdx <= num) {
        var blockSize:Int = java.lang.Math.pow(domain.size, num - varIdx).asInstanceOf[Int];
          
        pars = Element(value / blockSize + 1) :: pars;
        value = value % blockSize;
          
        varIdx += 1;
      }
        
      //image
      var image:Element = Element(varList(index).getValue());

      //append
      buffer = buffer ::: Pair(pars, image) :: Nil;
      index += 1;
    }
    buffer;
  }

  /*
   * Export this function as TPTP Annoted Formula
   */
  def export:AnnotedFormula = AnnotedFormula(name, FiFunctors, buildConjunction(getDefinedEntries))
  def buildConjunction(entries: List[Pair[List[Element],Element]]):Formula =  entries match {
    case Pair(params, value)::Nil => buildAtom(params,value)
    case Pair(params, value)::rest => And(buildAtom(params,value),buildConjunction(rest))
    case Nil => throw new IllegalArgumentException ()
  }
  def buildAtom(params:List[Element], value:Element):Formula = Atomic("=",Term(Symbol(name,false),(params map (p=> p.export)))::value.export::Nil)  
  
}

/**
 * Interpretable predicates
 */
case class PredicateRepr(domain:Domain, name:String, num:Int) extends InterpretationElementRepr() {
  /** Predicates evaluate to 1 or 2 (true or false) */
  def imageSize = 2;
  
  override def toString():String = "Predicate " + name + " (arity: " + num + ")"; 
  
  /**
   * Evaluate the current interpretation with a set of real argument values
   */
  def evaluate(parameter:List[Int]):Int = {
    //load the value in our interpretation
    var index = toIndex(parameter);
    varList(index).getValue();
  }
  
  /** Enumerate the interpretation */
    def getDefinedEntries():List[Pair[List[Element], boolean]] = {
      var buffer:List[Pair[List[Element],Boolean]] = Nil; 
      var index:Int = 0;
      while (index < varList.size) {
        //restore arguments        
        var pars:List[Element] = Nil;
        
        //now we have the pleasure of converting the index into the parameters
        var value = index; //base value
        var varIdx = 1;
        while (varIdx <= num) {
          var blockSize:Int = java.lang.Math.pow(domain.size, num - varIdx).asInstanceOf[Int];
          
          pars = Element(value / blockSize + 1) :: pars;
          value = value % blockSize;
          
          varIdx += 1;
        }
        
          
        //image
        var image:boolean = varList(index).getValue() == 1;

        //append
        buffer = buffer ::: Pair(pars, image) :: Nil;
        index += 1;
      }
      buffer;
    }
  
  /*
   * Export this function as TPTP Annoted Formula
   */
   def export:AnnotedFormula = AnnotedFormula(name, FiPredicates,buildConjunction(getDefinedEntries))     
   def buildConjunction(entries: List[Pair[List[Element],Boolean]]):Formula =  entries match {
     case Pair(params,value)::Nil => buildAtom(params,value)
     case Pair(params,value)::rest => And(buildAtom(params,value),buildConjunction(rest))
     case Nil => throw new IllegalArgumentException()
   }
   def buildAtom(params:List[Element], value:Boolean):Formula = if(value)Atomic(name,(params map (p=> p.export))) else Negation(Atomic(name,(params map (p=> p.export))))
  
}

/*
 * The built-in "=" predicate
 */
case class EqPredicate extends InterpretationElementRepr{

  def domain:Domain = new Domain(0);
  def name:String = "=";
  def num:Int = 2;
  def imageSize:Int = 2;
  
  /**
   * Evaluate the current interpretation with a set of real argument values
   */
  override def evaluate(parameter:List[Int]):Int = {
    if (parameter(0) == parameter(1)) {
      1
    } else {
      2 
    }
  }
  
  /** Increase the speed of modifications for this entry in the interpretation */
  override def increaseSpeed(pars:List[Int]):Unit = { }
  /** Lower the speed of modifications for this entry in the interpretation */
  override def decreaseSpeed(pars:List[Int]):Unit = { }
  
}

  
