package ch.epfl.lara.rmf;

import java.io.*;
import java.util.LinkedList;
import java.util.Hashtable;
import java.lang.Exception;

import ch.epfl.lara.tptp.SimpleTptpParserOutput;
import ch.epfl.lara.tptp.TptpLexer;
import ch.epfl.lara.tptp.TptpParser;
import ch.epfl.lara.tptp.TptpParserOutput;

public class TPTPParserTrial {

	public static final String INCLUDE_DIR = "C:/TPTP/";
	public static final String PROBLEM_DIR = "C:/TPTP/Problems";

	public static void main(String[] args) {

		try {
			if (args.length == 1) {
				String problemName = args[0];

				parseTPTPFile(PROBLEM_DIR,INCLUDE_DIR,problemName);

			}
		} catch (Exception e) {
			System.err.println("\nGENERAL exception: " + e);
		}

	}

	public static LinkedList<SimpleTptpParserOutput.TopLevelItem> parseTPTPFile(String problemDir, String includeDir,
			String problemName) throws Exception {
		String inFile = problemDir + "/" + problemName;

		/* the whole output of the parser */
		LinkedList<SimpleTptpParserOutput.TopLevelItem> allParsed = new LinkedList<SimpleTptpParserOutput.TopLevelItem>();

		/* create a parser output object */
		SimpleTptpParserOutput outputManager = new SimpleTptpParserOutput();

		/* prepare the symbol table for semantic checks */
		_signature.put(new String("="), new SymbolDescriptor(true, 2)); // "=" is a binary predicate 

		/* parse the input file and semantically check the results */
		parseAndCheck(inFile, includeDir, outputManager, allParsed, 0);

		System.out.println("File \"" + inFile + "\" is OK");

		System.out.println("Size: " + allParsed.size());
		return allParsed;
	}

	/** Parses the file <strong> fileName </strong> and checks the results semantically.
	 *  If <strong> includeDir </strong> != null, included files are parsed recursively
	 *  as soon as they appear.
	 */
	private static void parseAndCheck(String fileName, String includeDir,
			SimpleTptpParserOutput outputManager,
			LinkedList<SimpleTptpParserOutput.TopLevelItem> results,
			int recursionDepth) throws Exception {
		FileInputStream in = new FileInputStream(fileName);
		TptpLexer lexer = new TptpLexer(new DataInputStream(in));
		TptpParser parser = new TptpParser(lexer);

		for (SimpleTptpParserOutput.TopLevelItem item = (SimpleTptpParserOutput.TopLevelItem) parser
				.topLevelItem(outputManager); item != null; item = (SimpleTptpParserOutput.TopLevelItem) parser
				.topLevelItem(outputManager)) {
			if (includeDir != null
					&& item.getKind() == TptpParserOutput.TptpInput.Kind.Include) {
				if (recursionDepth >= 1024)
					throw new Exception(
							"Too many nested include directives (depth > 1024) "
									+ "in " + fileName + ", line "
									+ item.getLineNumber() + ".");

				String relativeIncludeFileName = ((SimpleTptpParserOutput.IncludeDirective) item)
						.getFileName();

				// The file name has to be truncated because it is quoted

				relativeIncludeFileName = relativeIncludeFileName.substring(1,
						relativeIncludeFileName.length() - 1);

				String includeFileName = includeDir + "/"
						+ relativeIncludeFileName;

				parseAndCheck(includeFileName, includeDir, outputManager,
						results, recursionDepth + 1);
			} else {
				/* Minimal semantic analysis: check that no symbol is used
				 both as a predicate and as a function or constant, and that 
				 no symbol is used with different arities.
				 checkSemantically(item) will throw 
				 java.lang.Exception(<error description>) if something is wrong 
				 with the item.
				 */
				checkSemantically(item, fileName);

				results.add(item);
			}
			;
		}
		;

		in.close();

	} // parseAndCheck(..)

	/** Checks whether <strong> item </strong> is semantically well-formed 
	 *  wrt the signature <strong> _signature </strong>. 
	 *  Throws java.lang.Exception(<error description>) if the use of some symbol
	 *  in <strong> item </strong> is inconsistent with its declaration in  
	 *  <strong> _signature </strong>. Side effect: new symbols are declared in 
	 *  <strong> _signature </strong>.
	 *  @param item != null
	 *  @param fileName file name to be used in error messages
	 */
	private static void checkSemantically(
			SimpleTptpParserOutput.TopLevelItem item, String fileName)
			throws Exception {
		assert item != null;

		_semanticCheckFileName = fileName;
		_semanticCheckLineNumber = item.getLineNumber();

		switch (item.getKind()) {
		case Formula:
			checkSemantically(((SimpleTptpParserOutput.AnnotatedFormula) item)
					.getFormula());
			break;
		case Clause:
			checkSemantically(((SimpleTptpParserOutput.AnnotatedClause) item)
					.getClause());
			break;
		case Include:
			break; // nothing to check
		}
	} // checkSemantically(SimpleTptpParserOutput.TopLevelItem item)

	/** Auxilliary for 
	 *  checkSemantically(SimpleTptpParserOutput.TopLevelItem item,String fileName):
	 *  checks whether <strong> formula </strong> is semantically well-formed 
	 *  wrt the signature <strong> _signature </strong>. Side effect: new symbols are 
	 *  declared in <strong> _signature </strong>.
	 */
	private static void checkSemantically(SimpleTptpParserOutput.Formula formula)
			throws Exception {

		switch (formula.getKind()) {
		case Atomic:
			checkSemantically((SimpleTptpParserOutput.Formula.Atomic) formula);
			break;
		case Negation:
			checkSemantically(((SimpleTptpParserOutput.Formula.Negation) formula)
					.getArgument());
			break;
		case Binary:
			checkSemantically(((SimpleTptpParserOutput.Formula.Binary) formula)
					.getLhs());
			checkSemantically(((SimpleTptpParserOutput.Formula.Binary) formula)
					.getRhs());
			break;
		case Quantified:
			checkSemantically(((SimpleTptpParserOutput.Formula.Quantified) formula)
					.getMatrix());
			break;
		}
	} // checkSemantically(SimpleTptpParserOutput.Formula formula)

	/** Auxilliary for 
	 *  checkSemantically(SimpleTptpParserOutput.TopLevelItem item,String fileName):
	 *  checks whether <strong> formula </strong> is semantically well-formed 
	 *  wrt the signature <strong> _signature </strong>. Side effect: new symbols are 
	 *  declared in <strong> _signature </strong>.
	 */
	private static void checkSemantically(SimpleTptpParserOutput.Clause clause)
			throws Exception {

		if (clause.getLiterals() != null) {

			for (SimpleTptpParserOutput.Literal lit : clause.getLiterals())
				checkSemantically(lit.getAtom());

		}
		;
	} // checkSemantically(SimpleTptpParserOutput.Clause clause)

	/** Auxilliary for 
	 *  checkSemantically(SimpleTptpParserOutput.TopLevelItem item,String fileName):
	 *  checks whether <strong> formula </strong> is semantically well-formed 
	 *  wrt the signature <strong> _signature </strong>. Side effect: new symbols are 
	 *  declared in <strong> _signature </strong>.
	 */
	private static void checkSemantically(
			SimpleTptpParserOutput.Formula.Atomic atom) throws Exception {

		String predicate = atom.getPredicate();

		int arity = atom.getNumberOfArguments();

		// Look up this symbol in the signature:
		SymbolDescriptor predicateDescriptor = _signature.get(predicate);

		if (predicateDescriptor == null) {
			// New symbol, has to be added to _signature
			_signature.put(predicate, new SymbolDescriptor(true, arity));
		} else {
			if (!predicateDescriptor.isPredicate) {
				throw new Exception("Semantic error in "
						+ _semanticCheckFileName + ", line "
						+ _semanticCheckLineNumber + ":" + "in atom " + atom
						+ ": predicate " + predicate
						+ " was used as a function elsewhere.");
			}
			;

			if (predicateDescriptor.arity != arity) {
				throw new Exception("Semantic error in "
						+ _semanticCheckFileName + ", line "
						+ _semanticCheckLineNumber + ":" + "in atom " + atom
						+ ": predicate " + predicate
						+ " was used with a different arity elsewhere.");
			}
			;
		}
		;

		// Check the arguments

		if (atom.getArguments() != null)
			for (SimpleTptpParserOutput.Term arg : atom.getArguments())
				checkSemantically(arg);

	} // checkSemantically(SimpleTptpParserOutput.Formula.Atomic atom)

	/** Auxilliary for 
	 *  checkSemantically(SimpleTptpParserOutput.TopLevelItem item,String fileName):
	 *  checks whether <strong> formula </strong> is semantically well-formed 
	 *  wrt the signature <strong> _signature </strong>. Side effect: new symbols are 
	 *  declared in <strong> _signature </strong>.
	 */
	private static void checkSemantically(SimpleTptpParserOutput.Term term)
			throws Exception {

		if (!term.getTopSymbol().isVariable()) {

			String function = term.getTopSymbol().getText();

			int arity = term.getNumberOfArguments();

			// Look up the function symbol in the signature
			SymbolDescriptor functionDescriptor = _signature.get(function);

			if (functionDescriptor == null) {
				// New symbol, has to be added to _signature
				_signature.put(function, new SymbolDescriptor(false, arity));
			} else {
				if (functionDescriptor.isPredicate) {
					throw new Exception("Semantic error in "
							+ _semanticCheckFileName + ", line "
							+ _semanticCheckLineNumber + ":" + "in term "
							+ term + ": function " + function
							+ " was used as a predicate elsewhere.");
				}
				;

				if (functionDescriptor.arity != arity) {
					throw new Exception("Semantic error in "
							+ _semanticCheckFileName + ", line "
							+ _semanticCheckLineNumber + ":" + "in term "
							+ term + ": function " + function
							+ " was used with a different arity elsewhere.");
				}
				;
			}
			;

			// Check the arguments

			if (term.getArguments() != null)
				for (SimpleTptpParserOutput.Term arg : term.getArguments())
					checkSemantically(arg);

		}
		; // if (!term.getTopSymbol().isVariable())
	} // checkSemantically(SimpleTptpParserOutput.Term term)

	/** Keeps information about the usage of a symbol: the category (predicate or function)
	 *  and the arity.
	 */
	private static class SymbolDescriptor {

		public SymbolDescriptor(boolean isPred, int ar) {
			isPredicate = isPred;
			arity = ar;
		}

		public boolean isPredicate;

		public int arity;

	} // class SymbolDescriptor

	/** Keeps descriptors of all symbols registered so far. */
	private static Hashtable<String, SymbolDescriptor> _signature = new Hashtable<String, SymbolDescriptor>();

	/** This is here to avoid threading this information through different
	 *  variants of checkSemantically(_).
	 */
	private static String _semanticCheckFileName;

	/** This is here to avoid threading this information through different
	 *  variants of checkSemantically(_).
	 */
	private static int _semanticCheckLineNumber;
	
	

} /* end of the class */
